from __future__ import annotations
import os
import sys

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
    overload,
    cast,
    TypeVar,
    Generic,
)

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from vowpal_wabbit_next import _core, Example

import numpy as np
import numpy.typing as npt
from vowpal_wabbit_next.labels import LabelType

from vowpal_wabbit_next.prediction_type import PredictionType

ScalarPrediction = float
ScalarsPrediction = List[float]
ActionScoresOrProbsPrediction = List[Tuple[int, float]]
DecisionScoresPrediction = List[List[Tuple[int, float]]]
MulticlassPrediction = int
MultilabelsPrediction = List[int]
PdfPrediction = List[Tuple[float, float, float]]
ActionPdfValuePrediction = Tuple[float, float]
ActiveMulticlassPrediction = Tuple[int, List[int]]
NoPrediction = type(None)

Prediction = Union[
    ScalarPrediction,
    ScalarsPrediction,
    ActionScoresOrProbsPrediction,
    DecisionScoresPrediction,
    MulticlassPrediction,
    MultilabelsPrediction,
    PdfPrediction,
    ActionPdfValuePrediction,
    ActiveMulticlassPrediction,
    NoPrediction,
]

MetricsDict = Dict[str, Union[int, float, str, bool, "MetricsDict"]]

DebugNode = _core.DebugNode

IsDebugT = TypeVar("IsDebugT")


class Workspace(Generic[IsDebugT]):
    @overload
    def __init__(
        self: Workspace[Literal[False]],
        args: List[str] = [],
        *,
        model_data: Optional[bytes] = None,
        record_feature_names: bool = False,
        record_metrics: bool = False,
        enable_debug_tree: Literal[False] = False,
    ):
        ...

    @overload
    def __init__(
        self: Workspace[Literal[True]],
        args: List[str] = [],
        *,
        model_data: Optional[bytes] = None,
        record_feature_names: bool = False,
        record_metrics: bool = False,
        enable_debug_tree: Literal[True] = True,
    ):
        ...

    @overload
    def __init__(
        self: Workspace[Literal[False]],
        *,
        _existing_workspace: Optional[_core.Workspace] = None,
    ):
        ...

    def __init__(
        self,
        args: List[str] = [],
        *,
        model_data: Optional[bytes] = None,
        record_feature_names: bool = False,
        record_metrics: bool = False,
        enable_debug_tree: bool = False,
        _existing_workspace: Optional[_core.Workspace] = None,
    ):
        """Main object used for making predictions and training a model.

        The VW library logs various things while running. There are two streams of logging exposed, which can be accessed via the standard Python logging interface.
        * `vowpal_wabbit_next.log` - VW's structured logging stream. If it outputs a warning it will be logged here.
        * `vowpal_wabbit_next.driver` - This is essentially the CLI driver output. This is rarely needed from Python.

        See the logging example below.

        Examples:
            Load a model from a file:

            >>> from vowpal_wabbit_next import Workspace
            >>> with open("model.bin", "rb") as f:
            ...     model_data = f.read()
            >>> workspace = Workspace(model_data=model_data)

            Create a workspace for training a contextual bandit with action dependent features model:

            >>> from vowpal_wabbit_next import Workspace
            >>> workspace = Workspace(["--cb_explore_adf"])

            Outputting structured logging messages from VW:

            >>> from vowpal_wabbit_next import Workspace
            >>> import logging
            >>> logging.basicConfig(level=logging.INFO)
            >>> logging.getLogger("vowpal_wabbit_next.log").setLevel("INFO")
            >>> workspace = Workspace()

        Args:
            args (List[str]): VowpalWabbit command line options for configuring the model. An overall list can be found `here <https://vowpalwabbit.org/docs/vowpal_wabbit/python/latest/command_line_args.html>`_. Options which affect the driver are not supported. For example:
                `--sort_features`, `--ngram`, `--feature_limit`, `--ignore`, `--extra_metrics`, `--dump_json_weights_experimental`
            model_data (Optional[bytes], optional): Bytes of a VW model to be loaded.
            record_invert_hash (bool, optional): If true, the invert hash will be recorded for each example. This is required to use :py:meth:`vowpal_wabbit_next.Workspace.json_weights`. This will slow down parsing and learn/predict.
            record_metrics (bool, optional): If true, reduction metrics will be enabled and can be fetched with :py:attr:`vowpal_wabbit_next.Workspace.metrics`
            enable_debug_tree (bool, optional): If true, debug information in the form of the computation tree will be emitted by :py:meth:`~vowpal_wabbit_next.learn_one`, :py:meth:`~vowpal_wabbit_next.predict_one` and :py:meth:`~vowpal_wabbit_next.predict_then_learn_one`. This will affect performance negatively. See :py:class:`~vowpal_wabbit_next.DebugNode` for more information.

                    .. warning::
                        This is an experimental feature.

            _existing_workspace (Optional[_core.Workspace], optional): This is for internal usage and should not be set by a user.
        """
        if _existing_workspace is not None:
            self._workspace = _existing_workspace
        else:
            self._workspace = _core.Workspace(
                args,
                model_data=model_data,
                record_feature_names=record_feature_names,
                record_metrics=record_metrics,
                debug=enable_debug_tree,
            )

    def _check_label(self, example: Union[Example, List[Example]]) -> None:
        if isinstance(example, list):
            for i, ex in enumerate(example):
                if self.label_type != ex.label_type:
                    raise ValueError(
                        f"Label type mismatch. Expected {self.label_type}, got {ex.label_type} for example {i}"
                    )
        else:
            if self.label_type != example.label_type:
                raise ValueError(
                    f"Label type mismatch. Expected {self.label_type}, got {example.label_type}"
                )

    @overload
    def predict_one(
        self: Workspace[Literal[True]], example: Union[Example, List[Example]]
    ) -> Tuple[Prediction, DebugNode]:
        ...

    @overload
    def predict_one(
        self: Workspace[Literal[False]], example: Union[Example, List[Example]]
    ) -> Prediction:
        ...

    def predict_one(
        self, example: Union[Example, List[Example]]
    ) -> Union[Prediction, Tuple[Prediction, DebugNode]]:
        """Make a single prediction.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser
            >>> workspace = Workspace()
            >>> parser = TextFormatParser(workspace)
            >>> workspace.learn_one(parser.parse_line("1.0 | price:.18 sqft:.15 age:.35 1976"))
            >>> workspace.predict_one(parser.parse_line("| price:.53 sqft:.32 age:.87 1924"))
            1.0

        Args:
            example (Union[Example, List[Example]]): Example to use for prediction. This should be a list if this workspace is :py:meth:`vowpal_wabbit_next.Workspace.multiline`, otherwise it is should be a single Example

        Returns:
            Prediction: Prediction produced by this example. Or, if `enable_debug_tree=True` was passed in the constructor then a tuple of prediction and :py:class:`~vowpal_wabbit_next.DebugNode` will be returned. The type corresponds to the :py:meth:`~vowpal_wabbit_next.Workspace.prediction_type` of the model. See :py:class:`~vowpal_wabbit_next.PredictionType` for the mapping to types.
        """
        self._check_label(example)

        if isinstance(example, Example):
            return self._workspace.predict_one(example._example)
        else:
            return self._workspace.predict_multi_ex_one([ex._example for ex in example])

    @overload
    def learn_one(
        self: Workspace[Literal[True]], example: Union[Example, List[Example]]
    ) -> List[DebugNode]:
        ...

    @overload
    def learn_one(
        self: Workspace[Literal[False]], example: Union[Example, List[Example]]
    ) -> None:
        ...

    def learn_one(
        self, example: Union[Example, List[Example]]
    ) -> Union[None, List[DebugNode]]:
        """Learn from one single example. Note, passing a list of examples here means the input is a multiline example, and not several individual examples. The label type of the example must match what is returned by :py:meth:`vowpal_wabbit_next.Workspace.label_type`.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser
            >>> workspace = Workspace()
            >>> parser = TextFormatParser(workspace)
            >>> workspace.learn_one(parser.parse_line("1.0 | price:.18 sqft:.15 age:.35 1976"))
            >>> workspace.predict_one(parser.parse_line("| price:.53 sqft:.32 age:.87 1924"))
            1.0

        Args:
            example (Union[Example, List[Example]]): Example to learn on.

        Returns:
            Union[None, DebugNode]: If `enable_debug_tree=True` was passed in the constructor then a :py:class:`~vowpal_wabbit_next.DebugNode` will be returned. Otherwise, None is returned.
        """
        self._check_label(example)

        if isinstance(example, Example):
            return self._workspace.learn_one(example._example)
        else:
            return self._workspace.learn_multi_ex_one([ex._example for ex in example])

    @overload
    def predict_then_learn_one(
        self: Workspace[Literal[True]], example: Union[Example, List[Example]]
    ) -> Tuple[Prediction, List[DebugNode]]:
        ...

    @overload
    def predict_then_learn_one(
        self: Workspace[Literal[False]], example: Union[Example, List[Example]]
    ) -> Prediction:
        ...

    def predict_then_learn_one(
        self: Workspace[Literal[True]], example: Union[Example, List[Example]]
    ) -> Union[Prediction, Tuple[Prediction, List[DebugNode]]]:
        """Make a prediction then learn from the example. This is potentially more efficient than a predict_one call followed by a learn_one call as the implementation is able to avoid duplicated work as long as the prediction is guaranteed to be from before learning.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser
            >>> workspace = Workspace()
            >>> parser = TextFormatParser(workspace)
            >>> workspace.predict_then_learn_one(parser.parse_line("1.0 | price:.18 sqft:.15 age:.35 1976"))
            0.0

        Args:
            example (Union[Example, List[Example]]): Example to use for prediction. This should be a list if this workspace is :py:meth:`vowpal_wabbit_next.Workspace.multiline`, otherwise it is should be a single Example

        Returns:
            Prediction: Prediction produced by this example.  Or, if `enable_debug_tree=True` was passed in the constructor then a tuple of prediction and :py:class:`~vowpal_wabbit_next.DebugNode` will be returned. The type corresponds to the :py:meth:`~vowpal_wabbit_next.Workspace.prediction_type` of the model. See :py:class:`~vowpal_wabbit_next.PredictionType` for the mapping to types.
        """
        self._check_label(example)

        if isinstance(example, Example):
            return self._workspace.predict_then_learn_one(example._example)
        else:
            return self._workspace.predict_then_learn_multi_ex_one(
                [ex._example for ex in example]
            )

    def end_pass(self) -> None:
        """Signal the end of a pass to the model."""
        self._workspace.end_pass()

    @property
    def prediction_type(self) -> PredictionType:
        """Based on the command line parameters used to setup VW a certain type of prediction is produced. See :py:class:`vowpal_wabbit_next.PredictionType` for the list of types and their corresponding Python type.

        Returns:
            PredictionType: The type of prediction this Workspace produces
        """
        return self._workspace.get_prediction_type()

    @property
    def label_type(self) -> LabelType:
        """Based on the command line parameters used to setup VW a certain label type is required.
        This can also be thought of as the type of problem being solved.

        Returns:
            LabelType: The type of label Examples must have to be used by this Workspace
        """
        return self._workspace.get_label_type()

    @property
    def multiline(self) -> bool:
        """Based on the command line parameters used to setup VW, the input to learn, predict or parsers expects either single Examples or lists of Examples.

        Returns:
            bool: True if this Workspace is configured as multiline, otherwise False
        """
        return self._workspace.get_is_multiline()

    @property
    def metrics(self) -> MetricsDict:
        """Metrics is a dictionary of names to values recorded internally by VW. There may be nested mappings. This is completely dependent on the command line parameters used to setup VW and the reductions enabled.

        Returns:
            MetricsDict: Dictionary of metrics

        Raises:
            ValueError: If the workspace is not configured to record metrics
        """
        return cast(MetricsDict, self._workspace.get_metrics())

    def serialize(self) -> bytes:
        """Serialize the current workspace as a VW model that can be loaded by the Workspace constructor, or command line tool.

        Returns:
            bytes: raw bytes of serialized Workspace
        """
        return self._workspace.serialize()

    def serialize_to_file(self, file_path: Union[str, os.PathLike[Any]]) -> None:
        """Serialize the current workspace as a VW model to a file."""
        return self._workspace.serialize_to_file(os.fspath(file_path))

    def weights(self) -> npt.NDArray[np.float32]:
        """Access to the weights of the model currently.

        This function returns a view of the weights and any changes to the returned array will be reflected in the model.

        There are 3 dimensions:

        * The feature index (aka weight index)
        * The index of the interleaved model, which should usually be 0
        * The weight itself and the extra state stored with the weight

        .. attention::
            Only dense weights are supported.

        .. warning::
            This is an experimental feature.

        Examples:
            >>> from vowpal_wabbit_next import Workspace
            >>> model = Workspace()
            >>> print(model.weights().shape)
            (262144, 1, 4)

        Returns:
            np.ndarray: Array of weights
        """
        return np.array(self._workspace.weights(), copy=False)

    def json_weights(
        self, *, include_feature_names: bool = False, include_online_state: bool = False
    ) -> str:
        """Debugging utility which dumps the weights in the model currently as a JSON string.

        .. warning::
            This is an experimental feature.

        Args:
            include_feature_names (bool, optional): Includes the feature names and interaction terms in the output. This requires the workspace to be configured to support it by passing `record_feature_names=True` to the constructor of :py:class:`vowpal_wabbit_next.Workspace`. Defaults to False.
            include_online_state (bool, optional): Includes extra save_resume state in the output.

        Returns:
            str: JSON string representing model weights

        Raises:
            ValueError: If the workspace is not configured to record feature names and include_feature_names is True
        """
        return self._workspace.json_weights(
            include_feature_names=include_feature_names,
            include_online_state=include_online_state,
        )

    def readable_model(self, *, include_feature_names: bool = False) -> str:
        """Equivalent to the `--readable_model` command line option. If `include_feature_names` is True, then the equivalent of `--invert_hash`.

        Args:
            include_feature_names (bool, optional): Includes the feature names and interaction terms in the output. This requires the workspace to be configured to support it by passing `record_feature_names=True` to the constructor of :py:class:`vowpal_wabbit_next.Workspace`. Defaults to False.

        Returns:
            str: JSON string representing model

        Raises:
            ValueError: If the workspace is not configured to record feature names and include_feature_names is True
        """
        return self._workspace.readable_model(
            include_feature_names=include_feature_names
        )

    def get_index_for_scalar_feature(
        self,
        feature_name: str,
        *,
        feature_value: Optional[str] = None,
        namespace_name: str = " ",
    ) -> int:
        """Calculate the model index for a given feature.

        The logic is rather complicated to work out an index. This function also takes into account index
        truncation caused by the index multiplier taking the index out of the standard weight space.

        .. warning::
            This is an experimental feature, the interface may change.

        Examples:
            >>> from vowpal_wabbit_next import Workspace
            >>> model = Workspace()
            >>> # Feature which looks like "|test thing" in text format
            >>> model.get_index_for_scalar_feature("thing", namespace_name="test")
            148099

        Args:
            feature_name (str): The name of the feature
            feature_value (Optional[str], optional): String value of feature. If passed chain hashing will be used. In text format this looks like `feature_name:feature_value`
            namespace_name (str, optional): Namespace of feature. Defaults to " " which is the default namespace.

        Returns:
            int: The index of the feature
        """
        return self._workspace.get_index_for_scalar_feature(
            feature_name=feature_name,
            feature_value=feature_value,
            namespace_name=namespace_name,
        )

    # TODO implement
    # def get_index_for_interacted_feature(
    #     self, terms: List[Tuple[str, Optional[str], str]]
    # ) -> int:
    #     """Calculate the hash for an interacted feature.

    #     .. warning::
    #         This is an experimental feature, the interface may change.

    #     Args:
    #         terms (List[Tuple[str, Optional[str], str]]): List of features which are interacted. (feature_name, feature_value, namespace_name)

    #     Returns:
    #         int: The index of the feature
    #     """
    #     raise NotImplementedError()



import typing

from vowpal_wabbit_next import _core, Workspace, Example
from types import TracebackType

T = typing.TypeVar("T")


class JsonFormatParser:
    def __init__(self, workspace: Workspace[T]):
        """Parse VW Json format examples.

        Args:
            workspace (Workspace): Workspace object used to configure this parser
        """
        self._workspace = workspace
        self._multi_line = workspace.multiline

    def parse_json(self, text: str) -> typing.Union[Example, typing.List[Example]]:
        '''Parse a single json object in json format.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, JsonFormatParser
            >>> workspace = Workspace()
            >>> parser = JsonFormatParser(workspace)
            >>> json_str = """
            ... {
            ...     "_label": -1.0,
            ...     "feat1": 0.5,
            ...     "feat2": 2,
            ... }
            ... """
            >>> example = parser.parse_json(json_str)

        Args:
            text (str): JSON string of input

        Returns:
            typing.Union[Example, typing.List[Example]]: Parsed example or list of examples
        '''
        result = _core._parse_line_json(self._workspace._workspace, text)
        if self._multi_line:
            return [
                Example(_existing_example=ex, _label_type=self._workspace.label_type)
                for ex in result
            ]
        else:
            if len(result) != 1:
                raise ValueError("Expected single example")
            return Example(
                _existing_example=result[0], _label_type=self._workspace.label_type
            )


JsonFormatReaderT = typing.TypeVar("JsonFormatReaderT", bound="JsonFormatReader")


# takes a file and uses a context manager to generate based on the contents of the file
class JsonFormatReader:
    def __init__(self, workspace: Workspace[T], file: typing.TextIO):
        """Read VW Json format examples from the given text file. This reader always produces lists of examples.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, JsonFormatReader
            >>> workspace = Workspace()
            >>> with open("data.txt", "r") as f:
            ...     with JsonFormatReader(workspace, f) as reader:
            ...         for example in reader:
            ...               workspace.predict_one(example)


        Args:
            workspace (Workspace): Workspace object used to configure this reader
            file (typing.BinaryIO): File to read from
        """
        self._parser = JsonFormatParser(workspace)
        self._workspace = workspace
        self._file = file
        if not self._workspace.multiline:
            raise ValueError("Must use a multiline Workspace for json format")

    def __enter__(self: JsonFormatReaderT) -> JsonFormatReaderT:
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[TracebackType],
    ) -> None:
        self._file.close()

    def __iter__(self) -> typing.Iterator[typing.Union[Example, typing.List[Example]]]:
        if self._workspace.multiline:
            for line in self._file:
                yield self._parser.parse_json(line.rstrip())



import vowpal_wabbit_next as vw
import pytest


def test_simple_label() -> None:
    example = vw.Example()
    assert example.get_label() is None
    example.set_label(vw.SimpleLabel(label=1.0))
    assert isinstance(example.get_label(), vw.SimpleLabel)
    assert example.get_label().label == 1.0
    assert example.get_label().weight == 1.0
    assert example.get_label().initial == 0.0

    example.set_label(vw.MulticlassLabel(label=2))
    assert example.get_label().label == 2


def test_simple_label_still_there_after_learn() -> None:
    model = vw.Workspace()
    parser = vw.TextFormatParser(model)

    example = parser.parse_line("1 | a b c")
    assert example.get_label().label == 1
    assert isinstance(example.get_label(), vw.SimpleLabel)
    model.learn_one(example)
    assert example.get_label().label == 1
    assert isinstance(example.get_label(), vw.SimpleLabel)


def test_multiclass_label() -> None:
    example = vw.Example()
    assert example.get_label() is None
    example.set_label(vw.MulticlassLabel(label=1))
    assert isinstance(example.get_label(), vw.MulticlassLabel)
    assert example.get_label().label == 1
    assert example.get_label().weight == 1.0
    example.set_label(None)
    assert example.get_label() is None


def test_cb_label() -> None:
    example = vw.Example()
    assert example.get_label() is None
    example.set_label(vw.CBLabel(label=(3.4, 1.2), weight=0.5))
    assert isinstance(example.get_label(), vw.CBLabel)
    assert example.get_label().label == pytest.approx((0, 3.4, 1.2))

    with pytest.raises(ValueError):
        example.set_label(vw.CBLabel(label=(3.4, 1.2), shared=True))

    example.set_label(vw.CBLabel(shared=True))
    assert isinstance(example.get_label(), vw.CBLabel)
    assert example.get_label().label is None
    assert example.get_label().shared == True

    example.set_label(None)
    assert example.get_label() is None


def test_cs_label() -> None:
    example = vw.Example()
    assert example.get_label() is None
    example.set_label(vw.CSLabel(costs=[(1, 1.2)]))
    assert isinstance(example.get_label(), vw.CSLabel)
    assert len(example.get_label().costs) == 1
    assert example.get_label().costs[0] == pytest.approx((1, 1.2))

    with pytest.raises(ValueError):
        example.set_label(vw.CSLabel(costs=[(1, 1.2)], shared=True))

    example.set_label(vw.CSLabel(shared=True))
    assert isinstance(example.get_label(), vw.CSLabel)
    assert example.get_label().costs is None
    assert example.get_label().shared == True

    example.set_label(None)
    assert example.get_label() is None


def test_cs_label_parsed() -> None:
    workspace = vw.Workspace(["--csoaa_ldf=mc"])
    parser = vw.TextFormatParser(workspace)
    shared_ex = parser.parse_line("shared | a b c")
    labeled_ex = parser.parse_line("1:0.5 | a b c")
    assert isinstance(shared_ex.get_label(), vw.CSLabel)
    assert isinstance(labeled_ex.get_label(), vw.CSLabel)

    assert shared_ex.get_label().costs is None
    assert shared_ex.get_label().shared == True

    assert len(labeled_ex.get_label().costs) == 1
    assert labeled_ex.get_label().costs[0] == pytest.approx((1, 0.5))
    assert labeled_ex.get_label().shared == False


def test_ccb_label() -> None:
    example = vw.Example()
    assert example.get_label() is None
    example.set_label(vw.CCBLabel(vw.CCBExampleType.Shared))
    assert isinstance(example.get_label(), vw.CCBLabel)
    assert example.get_label().example_type == vw.CCBExampleType.Shared

    with pytest.raises(ValueError):
        vw.CCBLabel(vw.CCBExampleType.Shared, outcome=(1.0, [(1, 0.8)]))

    with pytest.raises(ValueError):
        vw.CCBLabel(vw.CCBExampleType.Action, outcome=(1.0, [(1, 0.8)]))

    label = vw.CCBLabel(vw.CCBExampleType.Slot, outcome=(1.0, [(1, 0.8)]))
    assert label.outcome[0] == 1.0
    assert label.outcome[1][0] == pytest.approx((1, 0.8))

    label = vw.CCBLabel(vw.CCBExampleType.Slot, explicit_included_actions=[1, 2])
    assert label.explicit_included_actions == [1, 2]


def test_ccb_label_parsed() -> None:
    workspace = vw.Workspace(["--ccb_explore_adf"])
    parser = vw.TextFormatParser(workspace)
    shared_ex = parser.parse_line("ccb shared | a b c")
    assert isinstance(shared_ex.get_label(), vw.CCBLabel)
    assert shared_ex.get_label().example_type == vw.CCBExampleType.Shared
    assert shared_ex.get_label().outcome is None

    action_ex = parser.parse_line("ccb action | a b c")
    assert isinstance(action_ex.get_label(), vw.CCBLabel)
    assert action_ex.get_label().example_type == vw.CCBExampleType.Action
    assert action_ex.get_label().outcome is None

    slot_ex = parser.parse_line("ccb slot | a b c")
    assert isinstance(slot_ex.get_label(), vw.CCBLabel)
    assert slot_ex.get_label().example_type == vw.CCBExampleType.Slot
    assert slot_ex.get_label().outcome is None
    assert slot_ex.get_label().explicit_included_actions is None

    slot_ex = parser.parse_line("ccb slot 1:0.5:0.8,2:0.2 1,2,3 | a b c")
    assert slot_ex.get_label().example_type == vw.CCBExampleType.Slot
    assert slot_ex.get_label().outcome is not None
    cost, action_probs = slot_ex.get_label().outcome
    assert cost == pytest.approx(0.5)
    assert len(action_probs) == 2
    assert action_probs[0] == pytest.approx((1, 0.8))
    assert action_probs[1] == pytest.approx((2, 0.2))
    assert slot_ex.get_label().explicit_included_actions == [1, 2, 3]



import json
import vowpal_wabbit_next as vw
import pytest


def test_json_weights() -> None:
    model = vw.Workspace(["--noconstant"], record_feature_names=True)
    parser = vw.TextFormatParser(model)

    model.learn_one(parser.parse_line("1 | a b c"))

    json_weights = json.loads(model.json_weights(include_feature_names=True))
    assert len(json_weights["weights"]) == 3
    all_names = set()
    for weight in json_weights["weights"]:
        assert len(weight["terms"]) == 1
        all_names.add(weight["terms"][0]["name"])

    assert all_names == {"a", "b", "c"}


def test_json_weights_feat_name_without_constructor_enabled() -> None:
    model = vw.Workspace()
    with pytest.raises(RuntimeError):
        model.json_weights(include_feature_names=True)


def test_readable_model():
    model = vw.Workspace(record_feature_names=True)
    parser = vw.TextFormatParser(model)
    model.learn_one(parser.parse_line("1 | MY_SUPER_UNIQUE_FEATURE_NAME"))
    assert "MY_SUPER_UNIQUE_FEATURE_NAME" in model.readable_model(
        include_feature_names=True
    )


def test_readable_model_feat_name_without_constructor_enabled():
    model = vw.Workspace()
    with pytest.raises(RuntimeError):
        model.readable_model(include_feature_names=True)


def calc_depth(node, depth=1):
    if len(node.children) == 0:
        return depth
    return max(calc_depth(child, depth + 1) for child in node.children)


def test_debug_node_returned():
    workspace = vw.Workspace(enable_debug_tree=True)
    parser = vw.TextFormatParser(workspace)

    workspace.learn_one(parser.parse_line("0 | price:.23 sqft:.25 age:.05 2006"))
    workspace.learn_one(
        parser.parse_line("1 2 'second_house | price:.18 sqft:.15 age:.35 1976")
    )
    workspace.learn_one(
        parser.parse_line("0 1 0.5 'third_house | price:.053 sqft:.32 age:.87 1924")
    )

    prediction, dbg_node = workspace.predict_one(
        parser.parse_line("| price:0.25 sqft:0.8 age:0.1")
    )

    assert isinstance(dbg_node, vw.DebugNode)
    assert isinstance(dbg_node.output_prediction, type(prediction))
    # count_label, scorer, gd
    assert calc_depth(dbg_node) == 3


def test_automl_epsilon_decay_return_tree():
    workspace = vw.Workspace(
        ["--cb_explore_adf", "--automl=4", "--cover=4"], enable_debug_tree=True
    )
    parser = vw.TextFormatParser(workspace)

    ex = [
        parser.parse_line("shared | s_1"),
        parser.parse_line("0:0.1:0.25 | a:0.5 b:1"),
        parser.parse_line("| a:-1 b:-0.5"),
        parser.parse_line("| a:-2 b:-1"),
    ]

    dbg_node = workspace.learn_one(ex)
    assert isinstance(dbg_node, list)
    assert len(dbg_node) == 1
    assert calc_depth(dbg_node[0]) > 5



import io
import vowpal_wabbit_next as vw
from textwrap import dedent


def test_write_and_read_cache() -> None:
    text_input = io.StringIO(
        dedent(
            """1 | a b c
    0 | d e f
    """
        )
    )

    cache_buffer = io.BytesIO()
    read_buffer = io.BytesIO()
    workspace = vw.Workspace()
    write_counter = 0
    with vw.TextFormatReader(workspace, text_input) as reader:
        with vw.CacheFormatWriter(workspace, cache_buffer) as writer:
            for example in reader:
                writer.write_example(example)
                write_counter += 1

            read_buffer = io.BytesIO(cache_buffer.getvalue())

    read_counter = 0
    with vw.CacheFormatReader(workspace, read_buffer) as reader:
        for example in reader:
            assert isinstance(example, vw.Example)
            read_counter += 1

    assert write_counter == read_counter



from vowpal_wabbit_next import _core
from vowpal_wabbit_next.labels import (
    LabelType,
    SimpleLabel,
    MulticlassLabel,
    CBLabel,
    CSLabel,
    CCBLabel,
)
from typing import List, Optional, Union, Iterator

AllLabels = Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None]

FeatureGroupRef = _core.FeatureGroupRef


class Example:
    def __init__(
        self,
        *,
        _existing_example: Optional[_core.Example] = None,
        _label_type: LabelType = LabelType.NoLabel
    ):
        """A single VW example.

        Args:
            _existing_example (Optional[_core.Example], optional): This is an internal parameter and should not be used by end users.
            _label_type (LabelType, optional): This is an internal parameter and should not be used by end users.
        """
        if _existing_example is not None:
            if _label_type is None:
                raise ValueError(
                    "Must provide label type when using an existing example"
                )
            self._example = _existing_example
            self.label_type = _label_type
        else:
            self._example = _core.Example()
            self.label_type = LabelType.NoLabel

    def get_label(self) -> AllLabels:
        """Get the label of the example.

        Returns:
            AllLabels: The label of the example
        """
        return self._example._get_label(self.label_type)

    def set_label(self, label: AllLabels) -> None:
        """Set the label of the example.

        Args:
            label (AllLabels): The label to set

        Raises:
            ValueError: If the label type is not supported.
        """
        if isinstance(label, SimpleLabel):
            self.label_type = LabelType.Simple
        elif isinstance(label, MulticlassLabel):
            self.label_type = LabelType.Multiclass
        elif isinstance(label, CBLabel):
            self.label_type = LabelType.CB
        elif isinstance(label, CSLabel):
            self.label_type = LabelType.CS
        elif isinstance(label, CCBLabel):
            self.label_type = LabelType.CCB
        elif label is None:
            self.label_type = LabelType.NoLabel
        else:
            raise ValueError("Unsupported label type")

        self._example._set_label(label)

    @property
    def tag(self) -> str:
        """Get the tag of the example.

        Returns:
            str: The tag of the example
        """
        return self._example._get_tag()

    @tag.setter
    def tag(self, tag: str) -> None:
        """Set the tag of the example.

        Args:
            tag (str): The tag to set
        """
        self._example._set_tag(tag)

    @property
    def feat_group_indices(self) -> List[int]:
        """Get the populated feature groups for this example.

        Returns:
            List[int]: The populated feature groups for this example
        """
        return self._example._feat_group_indices

    def __iter__(self) -> Iterator[FeatureGroupRef]:
        """Iterate over the feature groups in this example.

        Examples:
            >>> for fg in example:
            >>>     print(fg.feat_group_index)
            >>>     print(fg.indices)
            >>>     print(fg.values)

        Yields:
            Iterator[FeatureGroupRef]: An iterator over the feature groups in this example
        """
        return self._example.__iter__()

    def __convert_key(self, key: Union[str, int]) -> int:
        if isinstance(key, str):
            return ord(key)
        else:
            # Key must be a valid byte
            if key < 0 or key >= 256:
                raise IndexError("Index out of range")
            return key

    def __getitem__(self, key: Union[str, int]) -> FeatureGroupRef:
        """Get a reference to a feature group. If it doesn't already exist it will be created.

        Args:
            key (Union[str, int]): Either the first character of namespace or index of namespace

        Examples:
            >>> print(example["d"].feat_group_index)
            >>> print(example["d"].indices)
            >>> print(example["d"].values)

        Returns:
            FeatureGroupRef: A reference to the feature group
        """
        return self._example.__getitem__(self.__convert_key(key))

    def __delitem__(self, key: Union[str, int]) -> None:
        """Delete a feature group.

        Args:
            key (Union[str, int]): Either the first character of namespace or index of namespace
        """
        return self._example.__delitem__(self.__convert_key(key))

    def __contains__(self, key: Union[str, int]) -> bool:
        """Check if a feature group exists.

        Args:
            key (Union[str, int]): Either the first character of namespace or index of namespace

        Returns:
            bool: True if the feature group exists, False otherwise
        """
        return self._example.__contains__(self.__convert_key(key))



from vowpal_wabbit_next import _core

LabelType = _core.LabelType
SimpleLabel = _core.SimpleLabel
MulticlassLabel = _core.MulticlassLabel
CBLabel = _core.CBLabel
CSLabel = _core.CSLabel
CCBLabel = _core.CCBLabel

CCBExampleType = _core.CCBExampleType



import vowpal_wabbit_next as vw
import pytest
import tempfile
from pathlib import Path


def test_save_and_load() -> None:
    model = vw.Workspace()
    parser = vw.TextFormatParser(model)
    train_example_input = "1 | a b c"
    test_example_input = "| b c"

    train_example = parser.parse_line(train_example_input)
    test_example = parser.parse_line(test_example_input)

    model.learn_one(train_example)
    pred1 = model.predict_one(test_example)
    data = model.serialize()

    model2 = vw.Workspace(model_data=data)
    parser2 = vw.TextFormatParser(model)
    test_example2 = parser2.parse_line(test_example_input)
    pred2 = model2.predict_one(test_example2)

    assert pred1 == pytest.approx(pred2)


def test_serialize_to_file_and_load() -> None:
    model = vw.Workspace()
    parser = vw.TextFormatParser(model)
    train_example_input = "1 | a b c"
    test_example_input = "| b c"

    train_example = parser.parse_line(train_example_input)
    test_example = parser.parse_line(test_example_input)

    model.learn_one(train_example)
    pred1 = model.predict_one(test_example)

    temp_dir = Path(tempfile.gettempdir())
    model_path = temp_dir / "test_serialize_to_file_and_load.vw"

    model.serialize_to_file(model_path)

    try:
        with open(model_path, "rb") as f:
            model2 = vw.Workspace(model_data=f.read())

        parser2 = vw.TextFormatParser(model)
        test_example2 = parser2.parse_line(test_example_input)
        pred2 = model2.predict_one(test_example2)

        assert pred1 == pytest.approx(pred2)
    finally:
        model_path.unlink()



import vowpal_wabbit_next as vw
import pytest
import numpy as np


def test_equivalent_models() -> None:
    model = vw.Workspace(["--invert_hash=unused"])
    parser = vw.TextFormatParser(model)

    model.learn_one(parser.parse_line("1 | a b c"))

    model_after_1_learn = vw.Workspace(model_data=model.serialize())

    model.learn_one(parser.parse_line("1 | d e f"))

    delta_of_second_learn = vw.calculate_delta(model_after_1_learn, model)
    model_after_1_learn_and_delta_of_second_learn_applied = vw.apply_delta(
        model_after_1_learn, delta_of_second_learn
    )

    test_example = "| d"
    assert model.predict_one(parser.parse_line(test_example)) == pytest.approx(
        model_after_1_learn_and_delta_of_second_learn_applied.predict_one(
            parser.parse_line(test_example)
        )
    )


def test_delta_aggregate_check_single():
    model_a = vw.Workspace([])
    text_parser = vw.TextFormatParser(model_a)

    line = "1 | a:1.1 b:0.3 c:-0.4"
    model_a.learn_one(text_parser.parse_line(line))
    weight_a = model_a.weights()[model_a.get_index_for_scalar_feature("a")][0][0]
    adaptive_a = model_a.weights()[model_a.get_index_for_scalar_feature("a")][0][1]

    model_b = vw.Workspace([])

    line = "1 | a:-1.7 b:0.9 c:1.3"
    model_b.learn_one(text_parser.parse_line(line))
    weight_b = model_b.weights()[model_b.get_index_for_scalar_feature("a")][0][0]
    adaptive_b = model_b.weights()[model_b.get_index_for_scalar_feature("a")][0][1]

    delta = vw.calculate_delta(vw.Workspace([]), model_a)
    another_delta = vw.calculate_delta(vw.Workspace([]), model_b)
    sum_deltas = vw.merge_deltas([delta, another_delta])
    new_model = vw.apply_delta(vw.Workspace([]), sum_deltas)

    ab_weight = new_model.weights()[new_model.get_index_for_scalar_feature("a")][0][0]
    ab_adaptive = new_model.weights()[new_model.get_index_for_scalar_feature("a")][0][1]

    adaptive_total = adaptive_a + adaptive_b
    assert ab_adaptive == pytest.approx(
        adaptive_a * adaptive_a / adaptive_total
        + adaptive_b * adaptive_b / adaptive_total
    )
    assert ab_weight == pytest.approx(
        weight_a * adaptive_a / adaptive_total + weight_b * adaptive_b / adaptive_total
    )


def test_delta_aggregate_check_all():
    model_a = vw.Workspace([])
    text_parser = vw.TextFormatParser(model_a)

    line = "1 | a:1.1 b:0.3 c:-0.4"
    model_a.learn_one(text_parser.parse_line(line))

    model_b = vw.Workspace([])

    line = "1 | a:-1.7 b:0.9 c:1.3"
    model_b.learn_one(text_parser.parse_line(line))

    delta = vw.calculate_delta(vw.Workspace([]), model_a)
    another_delta = vw.calculate_delta(vw.Workspace([]), model_b)
    sum_deltas = vw.merge_deltas([delta, another_delta])
    new_model = vw.apply_delta(vw.Workspace([]), sum_deltas)

    model_a_weights = model_a.weights()[:, :, 0]
    model_a_adaptives = model_a.weights()[:, :, 1]

    model_b_weights = model_b.weights()[:, :, 0]
    model_b_adaptives = model_b.weights()[:, :, 1]

    adaptive_sums = model_a_adaptives + model_b_adaptives
    # Use a masked_array to avoid divide by zero warnings for features with no data
    adaptive_sums = np.ma.masked_values(adaptive_sums, 0)

    # Ensure that corresponding indices are nonzero
    assert (
        np.ma.getmask(np.ma.masked_values(model_a_weights + model_b_weights, 0))
        == np.ma.getmask(adaptive_sums)
    ).all()

    # Bring back the zero values for masked away features
    reweighted_weights = (
        model_a_weights * model_a_adaptives / adaptive_sums
        + model_b_weights * model_b_adaptives / adaptive_sums
    ).filled(0)

    reweighted_adaptives = (
        model_a_adaptives * model_a_adaptives / adaptive_sums
        + model_b_adaptives * model_b_adaptives / adaptive_sums
    ).filled(0)

    assert np.allclose(new_model.weights()[:, :, 0], reweighted_weights)
    assert np.allclose(new_model.weights()[:, :, 1], reweighted_adaptives)



from io import StringIO
import vowpal_wabbit_next as vw
import logging


def test_logger_capture() -> None:
    log_stream = StringIO()
    driver_logger = logging.getLogger("vowpal_wabbit_next.driver")
    driver_logger.setLevel("INFO")
    stream_handler = logging.StreamHandler(stream=log_stream)
    driver_logger.addHandler(stream_handler)
    _ = vw.Workspace()
    driver_logger.removeHandler(stream_handler)
    assert log_stream.getvalue() != ""



import io
import vowpal_wabbit_next as vw
from textwrap import dedent


def test_simple_input() -> None:
    text_input = io.StringIO(
        dedent(
            """1 | a b c
        0 | d e f
    """
        )
    )
    workspace = vw.Workspace()
    counter = 0
    with vw.TextFormatReader(workspace, text_input) as reader:
        for example in reader:
            # perform assert example is not a list
            counter += 1
            assert isinstance(example, vw.Example)

    assert counter == 2


def test_cb_input() -> None:
    text_input = io.StringIO(
        dedent(
            """shared | s_1
        0:1.5:0.25 | a:0.5 b:1
        | a:-1 b:-0.5
        | a:-2 b:-1

        shared | s_1
        0:1.5:0.25 | a:0.5 b:1
        | a:-1 b:-0.5
        | a:-2 b:-1

        shared | s_1
        0:-1.5:0.5 | a:0.5 b:1
        | a:-1 b:-0.5
        | a:2 b:-1
    """
        )
    )
    workspace = vw.Workspace(["--cb_explore_adf"])
    counter = 0
    with vw.TextFormatReader(workspace, text_input) as reader:
        for example in reader:
            # perform assert example is not a list
            counter += 1
            assert isinstance(example, list)

    assert counter == 3



import io
import pytest
import vowpal_wabbit_next as vw


def test_example_creation() -> None:
    example = vw.Example()
    assert example.feat_group_indices == []

    feat_group = example["a"]
    assert example.feat_group_indices == [ord("a")]

    feat_group = example["a"]

    assert len(feat_group) == 0

    feat_group.push_feature(1, 1.0)
    assert len(feat_group) == 1

    with pytest.raises(Exception) as e_info:
        feat_group.truncate_to(4)

    feat_group.truncate_to(1)
    assert len(feat_group) == 1
    feat_group.truncate_to(0)
    assert len(feat_group) == 0

    feat_group.push_feature(1, 1.0)
    assert len(feat_group) == 1

    feat_group.push_many_features([2, 3], [2.0, 3.0])
    assert len(feat_group) == 3

    feat_group.push_feature(4, 4.0)
    assert len(feat_group) == 4

    feat_group.push_many_features([5, 6], [5.0, 6.0])
    assert len(feat_group) == 6

    feat_group.truncate_to(6)
    assert len(feat_group) == 6

    feat_group.truncate_to(4)
    assert len(feat_group) == 4


def test_example_creation_index() -> None:
    example = vw.Example()
    assert example.feat_group_indices == []

    with pytest.raises(Exception) as e_info:
        feat_group = example[256]

    with pytest.raises(Exception) as e_info:
        feat_group = example["test"]

    with pytest.raises(Exception) as e_info:
        feat_group = example[-56]

    assert example.feat_group_indices == []

    assert "a" not in example
    feat_group = example["a"]
    assert len(example.feat_group_indices) == 1
    assert "a" in example

    assert 0 not in example
    feat_group = example[0]
    assert len(example.feat_group_indices) == 2
    assert 0 in example

    del example[0]
    assert 0 not in example
    assert len(example.feat_group_indices) == 1

    del example["a"]
    assert "a" not in example
    assert len(example.feat_group_indices) == 0



import typing

from vowpal_wabbit_next import Example, Workspace, _core, TextFormatParser
from types import TracebackType

CacheFormatReaderT = typing.TypeVar("CacheFormatReaderT", bound="CacheFormatReader")

T = typing.TypeVar("T")


class CacheFormatReader:
    def __init__(self, workspace: Workspace[T], file: typing.BinaryIO):
        """Read VW examples in cache format from the given file.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser, CacheFormatWriter
            >>> workspace = Workspace()
            >>> with open("data.cache", "rb") as f:
            ...     with CacheFormatReader(workspace, f) as reader:
            ...         for example in reader:
            ...               workspace.predict_one(example)


        Args:
            workspace (Workspace): Workspace object used to configure this reader
            file (typing.BinaryIO): File to read from
        """
        self._workspace = workspace
        self._file = file
        self._reader = _core._CacheReader(self._workspace._workspace, self._file)

    def __enter__(self: CacheFormatReaderT) -> CacheFormatReaderT:
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[TracebackType],
    ) -> None:
        self._file.close()

    def __iter__(self) -> typing.Iterator[typing.Union[Example, typing.List[Example]]]:
        if self._workspace.multiline:
            so_far: typing.List[Example] = []
            # parse until we find a newline example
            next_example = self._reader._get_next()
            while next_example is not None:
                if next_example._is_newline():
                    if len(so_far) != 0:
                        yield so_far
                        so_far = []
                else:
                    so_far.append(
                        Example(
                            _existing_example=next_example,
                            _label_type=self._workspace.label_type,
                        )
                    )
                next_example = self._reader._get_next()
            if len(so_far) != 0:
                yield so_far
        else:
            next_example = self._reader._get_next()
            while next_example is not None:
                yield Example(
                    _existing_example=next_example,
                    _label_type=self._workspace.label_type,
                )
                next_example = self._reader._get_next()


CacheFormatWriterT = typing.TypeVar("CacheFormatWriterT", bound="CacheFormatWriter")


class CacheFormatWriter:
    def __init__(self, workspace: Workspace[T], file: typing.BinaryIO):
        """Creates a VW cache file.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser, CacheFormatWriter
            >>> workspace = Workspace()
            >>> parser = TextFormatParser(workspace)
            >>> with open("data.cache", "wb") as f:
            ...     with CacheFormatWriter(workspace, f) as writer:
            ...         writer.write_example(parser.parse_line("1.0 | price:.18 sqft:.15 age:.35 1976"))

        Args:
            workspace (Workspace): Workspace object used to configure this writer.
            file (typing.BinaryIO): File to write cache to
        """
        self._workspace = workspace
        self._file = file
        # TODO: workout a better way to handle this one...
        self._newline_example = TextFormatParser(workspace).parse_line("")
        _core._write_cache_header(self._workspace._workspace, self._file)

    def __enter__(self: CacheFormatWriterT) -> CacheFormatWriterT:
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[TracebackType],
    ) -> None:
        self._file.close()

    def write_example(
        self, example: typing.Union[Example, typing.List[Example]]
    ) -> None:
        """Write a single example to the cache file.

        Args:
            example (typing.Union[Example, typing.List[Example]]): Either a single or multiex to be written.
        """
        if isinstance(example, list):
            for e in example:
                _core._write_cache_example(
                    self._workspace._workspace, e._example, self._file
                )
            _core._write_cache_example(
                self._workspace._workspace, self._newline_example._example, self._file
            )
        else:
            _core._write_cache_example(
                self._workspace._workspace, example._example, self._file
            )



from typing import Generator, List, Optional, Tuple
from vowpal_wabbit_next import _core
import contextlib
from pathlib import Path
import os


# From: https://stackoverflow.com/questions/41742317/how-can-i-change-directory-with-python-pathlib
@contextlib.contextmanager
def _working_directory(path: Path) -> Generator[None, None, None]:
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)


class CLIError(Exception):
    def __init__(self, message: str, driver_output: str, log_output: List[str]):
        """Represents a failure when running the CLI. Exposes output for additional context.

        Args:
            message (str): Error message
            driver_output (str): Output of the VW driver
            log_output (List[str]): List of logs messages
        """
        super().__init__(message)
        self.driver_output = driver_output
        self.log_output = log_output


def run_cli_driver(
    args: List[str], *, onethread: bool = False, cwd: Optional[Path] = None
) -> Tuple[str, List[str]]:
    """Is the equivalent of running the VW command line tool with the given command line.

    There are a few differences:

    * Any input from stdin is not supported
    * The argfile input to command line is not supported
    * If any place in VW writes to stdout, stderr directly it is not captured. This means that `--version` and `--help` are not currently captured.

    .. warning::
        This is an experimental feature.

    Examples:
        >>> from vowpal_wabbit_next import run_cli_driver
        >>> driver_output, logs = run_cli_driver(["-d", "my_data.txt"])

        You can use `shlex` to split a command line:

        >>> from vowpal_wabbit_next import run_cli_driver
        >>> import shlex
        >>> driver_output, logs = run_cli_driver(shlex.split("-d my_data.txt"))

    Args:
        args (List[str]): Arguments to be passed to the command line driver
        onethread (bool, optional): Whether to use background thread for parsing. If False, a background thread is used for parsing. If True, no background threads are used and everything is done in the foreground of this call.
        cwd (Optional[Path], optional): The current working directory to use for the command line driver. If None, the current working directory is used.

    Raises:
        CLIError: If there is any error raised by execution.

    Returns:
        Tuple[str, List[str]]: driver output and log messages respectively as a tuple
    """
    with _working_directory(cwd) if cwd is not None else contextlib.nullcontext():
        error_info, driver_output, log_output = _core._run_cli_driver(
            args, onethread=onethread
        )
        if error_info is not None:
            raise CLIError(error_info, driver_output, log_output)

        return (driver_output, log_output)



import vowpal_wabbit_next as vw
import numpy as np


def test_learn() -> None:
    model: vw.Workspace = vw.Workspace()
    parser = vw.TextFormatParser(model)

    pred = model.predict_one(parser.parse_line("| a"))
    assert pred == 0.0
    model.learn_one(parser.parse_line("1 | a b c"))
    pred = model.predict_one(parser.parse_line("| a"))

    # Learn should result in a non-zero prediction.
    assert pred != 0

    idx = model.get_index_for_scalar_feature("a")

    assert model.weights()[idx][0][0] != 0


def test_predict_then_learn() -> None:
    model = vw.Workspace()
    parser = vw.TextFormatParser(model)

    pred = model.predict_then_learn_one(parser.parse_line("1 | a b c"))
    assert pred == 0.0
    pred = model.predict_one(parser.parse_line("| a"))

    # Learn should result in a non-zero prediction.
    assert pred != 0

    idx = model.get_index_for_scalar_feature("a")

    assert model.weights()[idx][0][0] != 0


def test_predict_then_learn_equivalent() -> None:
    model_learn = vw.Workspace()
    model_predict_and_learn = vw.Workspace()
    parser = vw.TextFormatParser(model_learn)

    model_learn.learn_one(parser.parse_line("1 | a b c"))
    model_learn.learn_one(parser.parse_line("2 | b d"))
    model_learn.learn_one(parser.parse_line("0.5 | b"))

    _ = model_predict_and_learn.predict_then_learn_one(parser.parse_line("1 | a b c"))
    _ = model_predict_and_learn.predict_then_learn_one(parser.parse_line("2 | b d"))
    _ = model_predict_and_learn.predict_then_learn_one(parser.parse_line("0.5 | b"))

    assert np.allclose(model_learn.weights(), model_predict_and_learn.weights())


# https://github.com/VowpalWabbit/py-vowpal-wabbit-next/issues/80
def test_oaa_learn_crash():
    dataset = [
        "2 |text three",
        "0 |text one",
        "1 |text two",
        "5 |text six",
        "3 |text four",
        "9 |text ten",
        "7 |text eight",
        "8 |text nine",
        "6 |text seven",
    ]

    workspace = vw.Workspace(
        ["--oaa=10", "--probabilities", "--loss_function=logistic"]
    )
    parser = vw.TextFormatParser(workspace)

    for line in dataset:
        workspace.learn_one(parser.parse_line(line))


def test_igl_predict_and_learn():
    dsjson_input = """{"_label_cost": 0, "_label_probability": 0.25, "_label_Action": 1, "_labelIndex": 0, "o": [{"v": {"v=click": 1}, "_definitely_bad": false}], "a": [0, 1, 2, 3], "c": {"User": {"user=Tom": 1, "time=afternoon" : 1}, "_multi": [{"Action": {"action=politics": 1}}, {"Action": {"action=sports": 1}}, {"Action": {"action=music": 1}}, {"Action": {"action=food": 1}}]}, "p": [0.25, 0.25, 0.25, 0.25]}"""
    model = vw.Workspace(
        ["--cb_explore_adf", "--coin", "--experimental_igl", "-q", "UA"]
    )

    parser = vw.DSJsonFormatParser(model)
    examples = parser.parse_json(dsjson_input)

    assert len(examples) == 6

    pred = model.predict_one(examples)
    assert all([p == 0.25 for _, p in pred])
    model.learn_one(examples)

    assert (
        model.weights()[
            model.get_index_for_scalar_feature("user=Tom", namespace_name="User")
        ][0][0]
        != 0
    )



import os
import vowpal_wabbit_next as vw
import pytest
import pathlib


def test_cli_produces_output() -> None:
    driver_output, log_output = vw.run_cli_driver(["-q::", "-qab"])
    assert len(driver_output) > 0
    # the -q should produce warnings
    assert len(log_output) > 0


def test_cli_onethread() -> None:
    driver_output, _ = vw.run_cli_driver([], onethread=True)
    assert len(driver_output) > 0


def test_cli_cwd() -> None:
    original_cwd = pathlib.Path(os.getcwd())

    # Can't find file in current dir
    with pytest.raises(vw.CLIError) as e_info:
        _, _ = vw.run_cli_driver(["--data=rcv1_small.dat"])

    data_dir = pathlib.Path(__file__).parent.resolve() / "data"
    # Can find file in data dir
    driver_output, _ = vw.run_cli_driver(["--data=rcv1_small.dat"], cwd=data_dir)

    # ensure the cwd is restored
    assert pathlib.Path(os.getcwd()) == original_cwd
    assert len(driver_output) > 0


def test_cli_cwd_restore_on_fail() -> None:
    original_cwd = pathlib.Path(os.getcwd())
    data_dir = pathlib.Path(__file__).parent.resolve() / "data"
    with pytest.raises(vw.CLIError) as e_info:
        driver_output, _ = vw.run_cli_driver(
            ["--data=rcv1_small.dat", "--bad_opt"], cwd=data_dir
        )

    assert pathlib.Path(os.getcwd()) == original_cwd


def test_cli_raises_error() -> None:
    with pytest.raises(vw.CLIError) as e_info:
        _, _ = vw.run_cli_driver(["--unknown_arg"])



from ._core import __version__, _vw_version, _vw_commit
from .example import Example
from .workspace import Workspace, DebugNode
from .text_format import TextFormatParser, TextFormatReader
from .json_format import JsonFormatParser, JsonFormatReader
from .dsjson_format import DSJsonFormatParser, DSJsonFormatReader
from .cache_format import CacheFormatWriter, CacheFormatReader
from .delta import ModelDelta, calculate_delta, apply_delta, merge_deltas
from .cli_driver import CLIError, run_cli_driver
from .prediction_type import PredictionType
from .labels import (
    LabelType,
    SimpleLabel,
    MulticlassLabel,
    CBLabel,
    CSLabel,
    CCBLabel,
    CCBExampleType,
)


VW_COMMIT: str = _vw_commit
"""Commit of VowpalWabbit that this package is built with"""
VW_VERSION: str = _vw_version
"""Version number of VowpalWabbit that this package is built with"""

__all__ = [
    "__version__",
    "apply_delta",
    "CacheFormatReader",
    "CacheFormatWriter",
    "calculate_delta",
    "CBLabel",
    "CLIError",
    "CSLabel",
    "CCBExampleType",
    "CCBLabel",
    "DebugNode",
    "DSJsonFormatParser",
    "DSJsonFormatReader",
    "Example",
    "JsonFormatParser",
    "JsonFormatReader",
    "LabelType",
    "merge_deltas",
    "ModelDelta",
    "MulticlassLabel",
    "PredictionType",
    "run_cli_driver",
    "SimpleLabel",
    "TextFormatParser",
    "TextFormatReader",
    "VW_COMMIT",
    "VW_VERSION",
    "Workspace",
]



from __future__ import annotations
import vowpal_wabbit_next._core
import typing
import numpy
_Shape = typing.Tuple[int, ...]

__all__ = [
    "CBLabel",
    "CCBExampleType",
    "CCBLabel",
    "CSLabel",
    "DebugNode",
    "DenseParameters",
    "Example",
    "FeatureGroupRef",
    "LabelType",
    "ModelDelta",
    "MulticlassLabel",
    "PredictionType",
    "SimpleLabel",
    "Workspace"
]


class CBLabel():
    def __init__(self, *, label: typing.Optional[typing.Union[typing.Tuple[float, float], typing.Tuple[int, float, float]]] = None, weight: float = 1.0, shared: bool = False) -> None: 
        """
        A label representing a contextual bandit problem.

        .. note::
          Currently the label can only contain 1 or 0 cb costs. There is a mode in vw for CB (non-adf) that allows for multiple cb_classes per example, but it is not currently accessible via direct label access. If creating examples/labels from parsed input it should still work as expected. If you need this feature, please open an issue on the github repo.

        Args:
          label (Optional[Union[Tuple[float, float], Tuple[int, float, float]]): This is (action, cost, probability). The same rules as VW apply for if the action can be left out of the tuple.
          weight (float): The weight of the example.
          shared (bool): Whether the example is shared. This is only used for ADF examples and must be the first example. There can only be one shared example per ADF example list.
        """
    def __repr__(self) -> str: ...
    @property
    def label(self) -> typing.Optional[typing.Tuple[int, float, float]]:
        """
            The label for the example. The format of the label is (action, cost, probability). If the action is not specified, it will be set to 0.

        :type: typing.Optional[typing.Tuple[int, float, float]]
        """
    @label.setter
    def label(self, arg1: typing.Optional[typing.Union[typing.Tuple[float, float], typing.Tuple[int, float, float]]]) -> None:
        """
        The label for the example. The format of the label is (action, cost, probability). If the action is not specified, it will be set to 0.
        """
    @property
    def shared(self) -> bool:
        """
            Whether the example is shared. This is only used for ADF examples and must be the first example. There can only be one shared example per ADF example list.

        :type: bool
        """
    @property
    def weight(self) -> float:
        """
            The weight of the example.

        :type: float
        """
    @weight.setter
    def weight(self, arg0: float) -> None:
        """
        The weight of the example.
        """
    pass
class CCBExampleType():
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    Action: vowpal_wabbit_next._core.CCBExampleType # value = <CCBExampleType.Action: 2>
    Shared: vowpal_wabbit_next._core.CCBExampleType # value = <CCBExampleType.Shared: 1>
    Slot: vowpal_wabbit_next._core.CCBExampleType # value = <CCBExampleType.Slot: 3>
    Unset: vowpal_wabbit_next._core.CCBExampleType # value = <CCBExampleType.Unset: 0>
    __members__: dict # value = {'Unset': <CCBExampleType.Unset: 0>, 'Shared': <CCBExampleType.Shared: 1>, 'Action': <CCBExampleType.Action: 2>, 'Slot': <CCBExampleType.Slot: 3>}
    pass
class CCBLabel():
    def __init__(self, type: CCBExampleType, *, outcome: typing.Optional[typing.Tuple[float, typing.List[typing.Tuple[int, float]]]] = None, explicit_included_actions: typing.Optional[typing.List[int]] = None) -> None: 
        """
        A label representing a conditional contextual bandit problem.

        Args:
          type (CCBExampleType): The type of the example. Unset is invalid.
          outcome (Optional[Tuple[float, List[Tuple[int, float]]]]): The outcome of the example. The format of the outcome is (cost, [(action, probability)]).
          explicit_included_actions (Optional[List[int]]): The list of actions explicitly included from the slot.
        """
    def __repr__(self) -> str: ...
    @property
    def example_type(self) -> CCBExampleType:
        """
            Whether the example represents the shared context.

        :type: CCBExampleType
        """
    @property
    def explicit_included_actions(self) -> typing.Optional[typing.List[int]]:
        """
            The list of actions explicitly excluded from the slot.

        :type: typing.Optional[typing.List[int]]
        """
    @property
    def outcome(self) -> typing.Optional[typing.Tuple[float, typing.List[typing.Tuple[int, float]]]]:
        """
            The outcome of the example. The format of the outcome is (cost, [(action, probability)]).

        :type: typing.Optional[typing.Tuple[float, typing.List[typing.Tuple[int, float]]]]
        """
    pass
class CSLabel():
    def __init__(self, *, costs: typing.Optional[typing.List[typing.Tuple[float, float]]] = None, shared: bool = False) -> None: 
        """
        A label representing a cost sensitive classification problem.

        Args:
          costs (Optional[List[Tuple[int, float]]]): List of classes and costs. If there is no label, this should be None.
          shared (bool): Whether the example represents the shared context
        """
    def __repr__(self) -> str: ...
    @property
    def costs(self) -> typing.Optional[typing.List[typing.Tuple[int, float]]]:
        """
            The costs for the example. The format of the costs is (class_index, cost).

        :type: typing.Optional[typing.List[typing.Tuple[int, float]]]
        """
    @costs.setter
    def costs(self, arg1: typing.List[typing.Tuple[int, float]]) -> None:
        """
        The costs for the example. The format of the costs is (class_index, cost).
        """
    @property
    def shared(self) -> bool:
        """
            Whether the example represents the shared context.

        :type: bool
        """
    pass
class DebugNode():
    """
    A node in the computation tree of a single learn/predict call. This represents the state of the example as it is entering a given reduction.

    .. warning::
      This is a highly experimental feature.
    """
    @property
    def children(self) -> typing.List[DebugNode]:
        """
        The child computations that this node processed. This represents traversal of the stack.

        :type: typing.List[DebugNode]
        """
    @property
    def function(self) -> str:
        """
        The function that was called on this reduction. Either 'learn' or 'predict'.

        :type: str
        """
    @property
    def input_labels(self) -> typing.Union[typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None], typing.List[typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None]]]:
        """
        The label that was passed into this reduction. Or, list of labels if this reduction is a multi-example reduction.

        :type: typing.Union[typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None], typing.List[typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None]]]
        """
    @property
    def interactions(self) -> typing.Union[typing.List[str], typing.List[typing.List[str]]]:
        """
        The interactions that were used to generate the features for this reduction. Or, list of interactions if this reduction is a multi-example reduction.

        :type: typing.Union[typing.List[str], typing.List[typing.List[str]]]
        """
    @property
    def is_multiline(self) -> bool:
        """
        Whether this reduction is a multi-example reduction.

        :type: bool
        """
    @property
    def name(self) -> str:
        """
        Name of the reduction being called.

        :type: str
        """
    @property
    def num_examples(self) -> int:
        """
        The number of examples that were processed by this reduction. This is always 1 for single example reductions.

        :type: int
        """
    @property
    def offset(self) -> typing.Union[int, typing.List[int]]:
        """
        The offset of the example. Or, list of offsets if this reduction is a multi-example reduction. This also includes the stride of the bottom learner.

        :type: typing.Union[int, typing.List[int]]
        """
    @property
    def output_prediction(self) -> typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None]:
        """
        The prediction that this reduction produced.

        :type: typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None]
        """
    @property
    def partial_prediction(self) -> typing.Union[float, typing.List[float]]:
        """
        The partial prediction on the example after this reduction ran. Or, list of partial predictions if this reduction is a multi-example reduction. This is generally only set by the bottom of the stack.

        :type: typing.Union[float, typing.List[float]]
        """
    @property
    def self_duration_ns(self) -> int:
        """
        The duration of this reduction in nanoseconds. It does not include time it takes to call children.

        :type: int
        """
    @property
    def total_duration_ns(self) -> int:
        """
        The duration of this reduction and all children in nanoseconds.

        :type: int
        """
    @property
    def updated_prediction(self) -> typing.Union[float, typing.List[float]]:
        """
        The partial prediction on the example after this reduction ran. Or, list of partial predictions if this reduction is a multi-example reduction. This is generally only set by the bottom of the stack.

        :type: typing.Union[float, typing.List[float]]
        """
    @property
    def weight(self) -> typing.Union[float, typing.List[float]]:
        """
        The weight of the example. Or, list of weights if this reduction is a multi-example reduction.

        :type: typing.Union[float, typing.List[float]]
        """
    pass
class DenseParameters():
    pass
class Example():
    def __contains__(self, arg0: int) -> bool: ...
    def __delitem__(self, arg0: int) -> None: ...
    def __getitem__(self, arg0: int) -> FeatureGroupRef: ...
    def __init__(self) -> None: ...
    def __iter__(self) -> typing.Iterator: ...
    def _get_label(self, arg0: LabelType) -> typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None]: ...
    def _get_tag(self) -> str: ...
    def _is_newline(self) -> bool: ...
    def _set_label(self, arg0: typing.Union[SimpleLabel, MulticlassLabel, CBLabel, CSLabel, CCBLabel, None]) -> None: ...
    def _set_tag(self, arg0: str) -> None: ...
    @property
    def _feat_group_indices(self) -> typing.List[int]:
        """
        :type: typing.List[int]
        """
    pass
class FeatureGroupRef():
    def __len__(self) -> int: ...
    def push_feature(self, index: int, value: float) -> None: 
        """
        Push a single feature into this group. This is an advanced function. Specifically, to ensure consistency with data that comes from parsers the index passed should incorporate the namespace hash.
        """
    def push_many_features(self, indices: numpy.ndarray[numpy.uint64], values: numpy.ndarray[numpy.float32]) -> None: 
        """
        Push many features into this group. This is an advanced function. Specifically, to ensure consistency with data that comes from parsers the index passed should incorporate the namespace hash.
        """
    def truncate_to(self, i: int) -> None: 
        """
        Truncate this feature group to the given size
        """
    @property
    def feat_group_index(self) -> int:
        """
        The index of the feature group. Since this is just the first letter of the namespace, multiple namespaces can map to the same group.

        :type: int
        """
    @property
    def indices(self) -> list:
        """
        Feature indices in this group.

        :type: list
        """
    @property
    def values(self) -> list:
        """
        Feature values in this group.

        :type: list
        """
    pass
class LabelType():
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    CB: vowpal_wabbit_next._core.LabelType # value = <LabelType.CB: 1>
    CBEval: vowpal_wabbit_next._core.LabelType # value = <LabelType.CBEval: 2>
    CBWithObservations: vowpal_wabbit_next._core.LabelType # value = <LabelType.CBWithObservations: 3>
    CCB: vowpal_wabbit_next._core.LabelType # value = <LabelType.CCB: 7>
    CS: vowpal_wabbit_next._core.LabelType # value = <LabelType.CS: 4>
    Continuous: vowpal_wabbit_next._core.LabelType # value = <LabelType.Continuous: 10>
    Multiclass: vowpal_wabbit_next._core.LabelType # value = <LabelType.Multiclass: 6>
    Multilabel: vowpal_wabbit_next._core.LabelType # value = <LabelType.Multilabel: 5>
    NoLabel: vowpal_wabbit_next._core.LabelType # value = <LabelType.NoLabel: 9>
    Simple: vowpal_wabbit_next._core.LabelType # value = <LabelType.Simple: 0>
    Slates: vowpal_wabbit_next._core.LabelType # value = <LabelType.Slates: 8>
    __members__: dict # value = {'Simple': <LabelType.Simple: 0>, 'CB': <LabelType.CB: 1>, 'CBEval': <LabelType.CBEval: 2>, 'CBWithObservations': <LabelType.CBWithObservations: 3>, 'CS': <LabelType.CS: 4>, 'Multilabel': <LabelType.Multilabel: 5>, 'Multiclass': <LabelType.Multiclass: 6>, 'CCB': <LabelType.CCB: 7>, 'Slates': <LabelType.Slates: 8>, 'NoLabel': <LabelType.NoLabel: 9>, 'Continuous': <LabelType.Continuous: 10>}
    pass
class ModelDelta():
    def __init__(self, model_data: bytes) -> None: ...
    def serialize(self) -> bytes: ...
    pass
class MulticlassLabel():
    def __init__(self, label: int, weight: float = 1.0) -> None: 
        """
        A label representing a multiclass classification problem.

        Args:
          label (int): The label.
          weight (float): The weight of the example.
        """
    @property
    def label(self) -> int:
        """
            The class of this label.

        :type: int
        """
    @label.setter
    def label(self, arg0: int) -> None:
        """
        The class of this label.
        """
    @property
    def weight(self) -> float:
        """
            The weight of this label.

        :type: float
        """
    @weight.setter
    def weight(self, arg0: float) -> None:
        """
        The weight of this label.
        """
    pass
class PredictionType():
    def __eq__(self, other: object) -> bool: ...
    def __getstate__(self) -> int: ...
    def __hash__(self) -> int: ...
    def __index__(self) -> int: ...
    def __init__(self, value: int) -> None: ...
    def __int__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __repr__(self) -> str: ...
    def __setstate__(self, state: int) -> None: ...
    @property
    def name(self) -> str:
        """
        :type: str
        """
    @property
    def value(self) -> int:
        """
        :type: int
        """
    ActionPdfValue: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.ActionPdfValue: 10>
    ActionProbs: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.ActionProbs: 4>
    ActionScores: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.ActionScores: 2>
    ActiveMulticlass: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.ActiveMulticlass: 11>
    DecisionProbs: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.DecisionProbs: 9>
    Multiclass: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Multiclass: 5>
    MulticlassProbs: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.MulticlassProbs: 8>
    Multilabels: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Multilabels: 6>
    NoPred: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.NoPred: 12>
    Pdf: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Pdf: 3>
    Prob: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Prob: 7>
    Scalar: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Scalar: 0>
    Scalars: vowpal_wabbit_next._core.PredictionType # value = <PredictionType.Scalars: 1>
    __members__: dict # value = {'Scalar': <PredictionType.Scalar: 0>, 'Scalars': <PredictionType.Scalars: 1>, 'ActionScores': <PredictionType.ActionScores: 2>, 'Pdf': <PredictionType.Pdf: 3>, 'ActionProbs': <PredictionType.ActionProbs: 4>, 'Multiclass': <PredictionType.Multiclass: 5>, 'Multilabels': <PredictionType.Multilabels: 6>, 'Prob': <PredictionType.Prob: 7>, 'MulticlassProbs': <PredictionType.MulticlassProbs: 8>, 'DecisionProbs': <PredictionType.DecisionProbs: 9>, 'ActionPdfValue': <PredictionType.ActionPdfValue: 10>, 'ActiveMulticlass': <PredictionType.ActiveMulticlass: 11>, 'NoPred': <PredictionType.NoPred: 12>}
    pass
class SimpleLabel():
    def __init__(self, label: float, weight: float = 1.0, initial: float = 0.0) -> None: 
        """
        A label representing a simple regression problem.

        Args:
          label (float): The label.
          weight (float): The weight of the example.
          initial (float): The initial value of the prediction.
        """
    def __repr__(self) -> str: ...
    @property
    def initial(self) -> float:
        """
            The initial value of the prediction.

        :type: float
        """
    @initial.setter
    def initial(self, arg0: float) -> None:
        """
        The initial value of the prediction.
        """
    @property
    def label(self) -> float:
        """
            The label.

        :type: float
        """
    @label.setter
    def label(self, arg0: float) -> None:
        """
        The label.
        """
    @property
    def weight(self) -> float:
        """
            The weight of this label.

        :type: float
        """
    @weight.setter
    def weight(self, arg0: float) -> None:
        """
        The weight of this label.
        """
    pass
class Workspace():
    def __init__(self, args: typing.List[str], *, model_data: typing.Optional[bytes] = None, record_feature_names: bool = False, record_metrics: bool = False, debug: bool = False) -> None: ...
    def end_pass(self) -> None: ...
    def get_index_for_scalar_feature(self, feature_name: str, feature_value: typing.Optional[str] = None, namespace_name: str = ' ') -> int: ...
    def get_is_multiline(self) -> bool: ...
    def get_label_type(self) -> LabelType: ...
    def get_metrics(self) -> dict: ...
    def get_prediction_type(self) -> PredictionType: ...
    def json_weights(self, *, include_feature_names: bool = False, include_online_state: bool = False) -> str: ...
    def learn_multi_ex_one(self, examples: typing.List[Example]) -> typing.Union[None, typing.List[DebugNode]]: ...
    def learn_one(self, examples: Example) -> typing.Union[None, typing.List[DebugNode]]: ...
    def predict_multi_ex_one(self, examples: typing.List[Example]) -> typing.Union[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.Tuple[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], DebugNode]]: ...
    def predict_one(self, examples: Example) -> typing.Union[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.Tuple[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], DebugNode]]: ...
    def predict_then_learn_multi_ex_one(self, examples: typing.List[Example]) -> typing.Union[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.Tuple[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.List[DebugNode]]]: ...
    def predict_then_learn_one(self, examples: Example) -> typing.Union[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.Tuple[typing.Union[float, typing.List[float], typing.List[typing.Tuple[int, float]], typing.List[typing.List[typing.Tuple[int, float]]], int, typing.List[int], typing.List[typing.Tuple[float, float, float]], typing.Tuple[float, float], typing.Tuple[int, typing.List[int]], None], typing.List[DebugNode]]]: ...
    def readable_model(self, *, include_feature_names: bool = False) -> str: ...
    def serialize(self) -> bytes: ...
    def serialize_to_file(self, arg0: str) -> None: ...
    def weights(self) -> DenseParameters: ...
    pass
class _CacheReader():
    def __init__(self, arg0: Workspace, arg1: object) -> None: ...
    def _get_next(self) -> typing.Optional[Example]: ...
    pass
def _apply_delta(base_workspace: Workspace, delta: ModelDelta) -> Workspace:
    pass
def _calculate_delta(base_workspace: Workspace, derived_workspace: Workspace) -> ModelDelta:
    pass
def _merge_deltas(deltas: typing.List[ModelDelta]) -> ModelDelta:
    pass
def _parse_line_dsjson(workspace: Workspace, line: str) -> typing.List[Example]:
    pass
def _parse_line_json(workspace: Workspace, line: str) -> typing.List[Example]:
    pass
def _parse_line_text(workspace: Workspace, line: str) -> Example:
    pass
def _run_cli_driver(args: typing.List[str], *, onethread: bool = False) -> typing.Tuple[typing.Optional[str], str, typing.List[str]]:
    pass
def _write_cache_example(workspace: Workspace, example: Example, file: object) -> None:
    pass
def _write_cache_header(workspace: Workspace, file: object) -> None:
    pass
__version__ = '0.6.0'
_vw_commit = '9db1f5f'
_vw_version = '9.9.0'



import typing

from vowpal_wabbit_next import _core, Workspace, Example
from types import TracebackType

T = typing.TypeVar("T")


class TextFormatParser:
    def __init__(self, workspace: Workspace[T]):
        """Parse VW text format examples.

        Args:
            workspace (Workspace): Workspace object used to configure this parser
        """
        self._workspace = workspace

    def parse_line(self, text: str) -> Example:
        """Parse a single line.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatParser
            >>> workspace = Workspace()
            >>> parser = TextFormatParser(workspace)
            >>> example = parser.parse_line("1.0 | price:.18 sqft:.15 age:.35 1976")

        Args:
            text (str): Text to parse

        Returns:
            Example: Parsed example
        """
        return Example(
            _existing_example=_core._parse_line_text(self._workspace._workspace, text),
            _label_type=self._workspace.label_type,
        )


TextFormatReaderT = typing.TypeVar("TextFormatReaderT", bound="TextFormatReader")


# takes a file and uses a context manager to generate based on the contents of the file
class TextFormatReader:
    def __init__(self, workspace: Workspace[T], file: typing.TextIO):
        """Read VW text format examples from the given text file. This reader produces either single Examples or List[Example] based on if the given workspace is multiline or not.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, TextFormatReader
            >>> workspace = Workspace()
            >>> with open("data.txt", "r") as f:
            ...     with TextFormatReader(workspace, f) as reader:
            ...         for example in reader:
            ...               workspace.predict_one(example)


        Args:
            workspace (Workspace): Workspace object used to configure this reader
            file (typing.BinaryIO): File to read from
        """
        self._parser = TextFormatParser(workspace)
        self._workspace = workspace
        self._file = file

    def __enter__(self: TextFormatReaderT) -> TextFormatReaderT:
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[TracebackType],
    ) -> None:
        self._file.close()

    def __iter__(self) -> typing.Iterator[typing.Union[Example, typing.List[Example]]]:
        if self._workspace.multiline:
            so_far: typing.List[Example] = []
            # parse until we find a newline example
            for line in self._file:
                ex = self._parser.parse_line(line)
                if ex._example._is_newline():
                    if len(so_far) != 0:
                        yield so_far
                        so_far = []
                else:
                    so_far.append(ex)
            if len(so_far) != 0:
                yield so_far
        else:
            for line in self._file:
                yield self._parser.parse_line(line)



import io
import vowpal_wabbit_next as vw
import pytest


def test_single() -> None:
    text_input = """{"_label_cost":-0.0,"_label_probability":0.05000000074505806,"_label_Action":4,"_labelIndex":3,"o":[{"v":0.0,"EventId":"13118d9b4c114f8485d9dec417e3aefe","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:29.2460000Z","Version":"1","EventId":"13118d9b4c114f8485d9dec417e3aefe","a":[4,2,1,3],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Sunny","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.05,0.05,0.05,0.85],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}"""
    workspace = vw.Workspace(["--cb_explore_adf"])
    parser = vw.DSJsonFormatParser(workspace)
    result = parser.parse_json(text_input)
    assert len(result) == 5
    assert isinstance(result, list)
    assert result[0].get_label().shared == True
    assert result[0].get_label().label is None
    assert result[1].get_label().shared == False
    assert result[1].get_label().label is None
    assert result[2].get_label().label is None
    assert result[3].get_label().label is None
    assert result[4].get_label().label is not None
    assert result[4].get_label().label == pytest.approx((4, 0.0, 0.05))


def test_single_formatted() -> None:
    text_input = """{
  "_label_cost": -1.0,
  "_label_probability": 0.8500000238418579,
  "_label_Action": 1,
  "_labelIndex": 0,
  "o": [
    {
      "v": 1.0,
      "EventId": "bf50a49c34b74937a81e8d6fc95faa99",
      "ActionTaken": false
    }
  ],
  "Timestamp": "2021-02-04T16:31:29.9430000Z",
  "Version": "1",
  "EventId": "bf50a49c34b74937a81e8d6fc95faa99",
  "a": [1, 3, 2, 4],
  "c": {
    "FromUrl": [
      { "timeofday": "Evening", "weather": "Snowy", "name": "Alice" }
    ],
    "_multi": [
      {
        "_tag": "Cappucino",
        "i": { "constant": 1, "id": "Cappucino" },
        "j": [
          {
            "type": "hot",
            "origin": "kenya",
            "organic": "yes",
            "roast": "dark"
          }
        ]
      },
      {
        "_tag": "Cold brew",
        "i": { "constant": 1, "id": "Cold brew" },
        "j": [
          {
            "type": "cold",
            "origin": "brazil",
            "organic": "yes",
            "roast": "light"
          }
        ]
      },
      {
        "_tag": "Iced mocha",
        "i": { "constant": 1, "id": "Iced mocha" },
        "j": [
          {
            "type": "cold",
            "origin": "ethiopia",
            "organic": "no",
            "roast": "light"
          }
        ]
      },
      {
        "_tag": "Latte",
        "i": { "constant": 1, "id": "Latte" },
        "j": [
          {
            "type": "hot",
            "origin": "brazil",
            "organic": "no",
            "roast": "dark"
          }
        ]
      }
    ]
  },
  "p": [0.85, 0.05, 0.05, 0.05],
  "VWState": {
    "m": "ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"
  },
  "_original_label_cost": -1.0
}
    """
    workspace = vw.Workspace(["--cb_explore_adf"])
    parser = vw.DSJsonFormatParser(workspace)
    result = parser.parse_json(text_input)
    assert len(result) == 5
    assert isinstance(result, list)


def test_reader() -> None:
    text_input = io.StringIO(
        """{"_label_cost":-0.0,"_label_probability":0.05000000074505806,"_label_Action":4,"_labelIndex":3,"o":[{"v":0.0,"EventId":"13118d9b4c114f8485d9dec417e3aefe","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:29.2460000Z","Version":"1","EventId":"13118d9b4c114f8485d9dec417e3aefe","a":[4,2,1,3],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Sunny","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.05,0.05,0.05,0.85],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}
{"_label_cost":-1.0,"_label_probability":0.8500000238418579,"_label_Action":1,"_labelIndex":0,"o":[{"v":1.0,"EventId":"bf50a49c34b74937a81e8d6fc95faa99","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:29.9430000Z","Version":"1","EventId":"bf50a49c34b74937a81e8d6fc95faa99","a":[1,3,2,4],"c":{"FromUrl":[{"timeofday":"Evening","weather":"Snowy","name":"Alice"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-1.0}
{"_label_cost":-0.0,"_label_probability":0.05000000074505806,"_label_Action":2,"_labelIndex":1,"o":[{"v":0.0,"EventId":"a4460f60db7a4088a82a30728e9d4bab","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:30.6530000Z","Version":"1","EventId":"a4460f60db7a4088a82a30728e9d4bab","a":[2,3,4,1],"c":{"FromUrl":[{"timeofday":"Morning","weather":"Sunny","name":"Bob"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.05,0.85,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}
{"_label_cost":-1.0,"_label_probability":0.8500000238418579,"_label_Action":1,"_labelIndex":0,"o":[{"v":1.0,"EventId":"d8b78f0f016746d6b72e2889f0f58655","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:31.1620000Z","Version":"1","EventId":"d8b78f0f016746d6b72e2889f0f58655","a":[1,4,3,2],"c":{"FromUrl":[{"timeofday":"Morning","weather":"Snowy","name":"Alice"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-1.0}
{"_label_cost":-0.0,"_label_probability":0.8500000238418579,"_label_Action":1,"_labelIndex":0,"o":[{"v":0.0,"EventId":"1d9e9975f2274242bcc2d109674e09ac","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:32.3270000Z","Version":"1","EventId":"1d9e9975f2274242bcc2d109674e09ac","a":[1,2,3,4],"c":{"FromUrl":[{"timeofday":"Morning","weather":"Sunny","name":"Alice"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}
{"_label_cost":-0.0,"_label_probability":0.8500000238418579,"_label_Action":4,"_labelIndex":3,"o":[{"v":0.0,"EventId":"ea6be41bc1b04eb89befdce03f3fe3a1","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:32.9270000Z","Version":"1","EventId":"ea6be41bc1b04eb89befdce03f3fe3a1","a":[4,1,3,2],"c":{"FromUrl":[{"timeofday":"Evening","weather":"Rainy","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}
{"_label_cost":-0.0,"_label_probability":0.8500000238418579,"_label_Action":3,"_labelIndex":2,"o":[{"v":0.0,"EventId":"2c53de8bf44749789d205b101025f7af","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:33.6460000Z","Version":"1","EventId":"2c53de8bf44749789d205b101025f7af","a":[3,2,1,4],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Sunny","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}
{"_label_cost":-1.0,"_label_probability":0.8500000238418579,"_label_Action":3,"_labelIndex":2,"o":[{"v":1.0,"EventId":"cea16d42810c41429e6a2c468769f70c","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:34.1620000Z","Version":"1","EventId":"cea16d42810c41429e6a2c468769f70c","a":[3,1,4,2],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Snowy","name":"Bob"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-1.0}
{"_label_cost":-1.0,"_label_probability":0.8500000238418579,"_label_Action":1,"_labelIndex":0,"o":[{"v":1.0,"EventId":"2dfd60f5f47a451fbe7f674afc5cafd7","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:35.3180000Z","Version":"1","EventId":"2dfd60f5f47a451fbe7f674afc5cafd7","a":[1,3,4,2],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Snowy","name":"Dave"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-1.0}
{"_label_cost":-1.0,"_label_probability":0.8500000238418579,"_label_Action":4,"_labelIndex":3,"o":[{"v":1.0,"EventId":"9c21eb678a234fb781d9b57541a8ee4d","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:35.9220000Z","Version":"1","EventId":"9c21eb678a234fb781d9b57541a8ee4d","a":[4,1,3,2],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Rainy","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.85,0.05,0.05,0.05],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-1.0}"""
    )
    workspace = vw.Workspace(["--cb_explore_adf"])
    counter = 0
    with vw.DSJsonFormatReader(workspace, text_input) as reader:
        for example in reader:
            # perform assert example is not a list
            counter += 1
            assert isinstance(example, list)
            assert len(example) == 5

    assert counter == 10



import vowpal_wabbit_next as vw
import pytest


def test_get_metric() -> None:
    text_input = """{"_label_cost":-0.0,"_label_probability":0.05000000074505806,"_label_Action":4,"_labelIndex":3,"o":[{"v":0.0,"EventId":"13118d9b4c114f8485d9dec417e3aefe","ActionTaken":false}],"Timestamp":"2021-02-04T16:31:29.2460000Z","Version":"1","EventId":"13118d9b4c114f8485d9dec417e3aefe","a":[4,2,1,3],"c":{"FromUrl":[{"timeofday":"Afternoon","weather":"Sunny","name":"Cathy"}],"_multi":[{"_tag":"Cappucino","i":{"constant":1,"id":"Cappucino"},"j":[{"type":"hot","origin":"kenya","organic":"yes","roast":"dark"}]},{"_tag":"Cold brew","i":{"constant":1,"id":"Cold brew"},"j":[{"type":"cold","origin":"brazil","organic":"yes","roast":"light"}]},{"_tag":"Iced mocha","i":{"constant":1,"id":"Iced mocha"},"j":[{"type":"cold","origin":"ethiopia","organic":"no","roast":"light"}]},{"_tag":"Latte","i":{"constant":1,"id":"Latte"},"j":[{"type":"hot","origin":"brazil","organic":"no","roast":"dark"}]}]},"p":[0.05,0.05,0.05,0.85],"VWState":{"m":"ff0744c1aa494e1ab39ba0c78d048146/550c12cbd3aa47f09fbed3387fb9c6ec"},"_original_label_cost":-0.0}"""
    workspace = vw.Workspace(["--cb_explore_adf"], record_metrics=True)
    parser = vw.DSJsonFormatParser(workspace)
    workspace.learn_one(parser.parse_json(text_input))
    metrics = workspace.metrics
    assert len(metrics) > 0
    assert "sfm_count_learn_example_with_shared" in metrics


def test_get_metric_not_enabled() -> None:
    workspace = vw.Workspace(["--cb_explore_adf"])

    with pytest.raises(RuntimeError):
        workspace.metrics



import sys
from typing import List, Optional, TypeVar

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from vowpal_wabbit_next import _core, Workspace


class ModelDelta:
    def __init__(
        self, data: bytes, *, _existing_model_delta: Optional[_core.ModelDelta] = None
    ):
        """A delta between two VW models.

        The standard way to create one is with :py:meth:`vowpal_wabbit_next.calculate_delta`.

        Args:
            data (bytes): Bytes of a previously serialized ModelDelta to be used for loading
            _existing_model_delta (Optional[_core.ModelDelta], optional): This is an internal parameter and should not be used by end users.
        """
        if _existing_model_delta is not None:
            self._model_delta = _existing_model_delta
        else:
            self._model_delta = _core.ModelDelta(data)

    def serialize(self) -> bytes:
        """Serialize the delta.

        Returns:
            bytes: The serialized delta.
        """
        return self._model_delta.serialize()


T = TypeVar("T")


def calculate_delta(
    base_model: Workspace[T], derived_model: Workspace[T]
) -> ModelDelta:
    """Produce a delta between two existing models.

    Args:
        base_model (Workspace): The base of the model
        derived_model (Workspace): The model produced from further training of base_model

    Returns:
        ModelDelta: The delta between the models.
    """
    return ModelDelta(
        bytes(),
        _existing_model_delta=_core._calculate_delta(
            base_model._workspace, derived_model._workspace
        ),
    )


def apply_delta(
    model: Workspace[Literal[False]], delta: ModelDelta
) -> Workspace[Literal[False]]:
    """Apply the delta to the model.

    Args:
        model (Workspace): The model to apply the delta to
        delta (ModelDelta): The delta to apply

    Returns:
        Workspace: The new model
    """
    return Workspace(
        _existing_workspace=_core._apply_delta(model._workspace, delta._model_delta)
    )


def merge_deltas(deltas: List[ModelDelta]) -> ModelDelta:
    """Merge a list of deltas into a single delta.

    Args:
        deltas (List[ModelDelta]): The deltas to merge. All deltas should come from the same base model.

    Returns:
        ModelDelta: The merged delta.
    """
    return ModelDelta(
        bytes(),
        _existing_model_delta=_core._merge_deltas([x._model_delta for x in deltas]),
    )



import sys
import pathlib

try:
    from skbuild import setup
except ImportError:
    print(
        "Please update pip, you need pip 10 or greater,\n"
        " or you need to install the PEP 518 requirements in pyproject.toml yourself",
        file=sys.stderr,
    )
    raise

from setuptools import find_packages

with open(pathlib.Path(__file__).parent.resolve() / "README.md", encoding="utf-8") as f:
    long_description = f.read()

with open(pathlib.Path(__file__).parent.resolve() / "version.txt", "r", encoding="utf-8") as version_file:
    version = version_file.read().strip()


setup(
    name="vowpal-wabbit-next",
    version=version,
    description="Experimental python bindings for VowpalWabbit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="VowpalWabbit",
    license="BSD-3-Clause",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"vowpal_wabbit_next": ["py.typed", "_core.pyi"]},
    cmake_install_dir="src/vowpal_wabbit_next",
    install_requires=["numpy", "typing_extensions; python_version=='3.7'", ],
    python_requires=">=3.7",
)



import vowpal_wabbit_next as vw


def test_single() -> None:
    workspace = vw.Workspace()
    parser = vw.JsonFormatParser(workspace)
    json_str = """{"_label":-1.0,"feat1":0.5,"feat2":2}"""
    example = parser.parse_json(json_str)
    assert isinstance(example, vw.Example)
    assert isinstance(example.get_label(), vw.SimpleLabel)


def test_multi() -> None:
    workspace = vw.Workspace(["--cb_adf"])
    parser = vw.JsonFormatParser(workspace)
    json_str = """{"s_":"1","s_":"2","_labelIndex":0,"_label_Action":0,"_label_Cost":1,"_label_Probability":0.5,"_multi":[{"a_":"1","b_":"1","c_":"1"}, {"a_":"2","b_":"2","c_":"2"},{"a_":"3","b_":"3","c_":"3"}]}"""
    example = parser.parse_json(json_str)
    assert isinstance(example, list)
    assert isinstance(example[0].get_label(), vw.CBLabel)



import vowpal_wabbit_next as vw


def test_get_scalar_index() -> None:
    model = vw.Workspace()

    assert 69985 == model.get_index_for_scalar_feature(
        "val", feature_value="test", namespace_name="another"
    )
    assert 92594 == model.get_index_for_scalar_feature("a")
    assert 148099 == model.get_index_for_scalar_feature("thing", namespace_name="test")
    assert 163331 == model.get_index_for_scalar_feature("b")

    model_with_wpp = vw.Workspace(["--automl=4", "--cb_adf"])
    # This hash causes a truncation since the wpp multiplies beyond the end of the weight array
    # We are ensuring the wrap around is calculated correctly.
    assert 45701 == model_with_wpp.get_index_for_scalar_feature(
        "feature", namespace_name="namespace"
    )


def test_get_weight_interleaved_models() -> None:
    model = vw.Workspace(["--automl=4", "--cb_adf"])
    parser = vw.TextFormatParser(model)

    example = [
        parser.parse_line("shared | s_1"),
        parser.parse_line("0:1.5:0.25 |namespace feature a:0.5 b:1"),
        parser.parse_line("| a:-1 b:-0.5"),
        parser.parse_line("| a:-2 b:-1"),
    ]
    model.learn_one(example)

    assert (
        model.weights()[
            model.get_index_for_scalar_feature("feature", namespace_name="namespace")
        ][0][0]
        != 0
    )
    assert (
        model.weights()[
            model.get_index_for_scalar_feature("feature", namespace_name="namespace")
        ][1][0]
        != 0
    )
    assert (
        model.weights()[
            model.get_index_for_scalar_feature("feature", namespace_name="namespace")
        ][2][0]
        != 0
    )
    assert (
        model.weights()[
            model.get_index_for_scalar_feature("feature", namespace_name="namespace")
        ][3][0]
        != 0
    )
    assert (
        model.weights()[model.get_index_for_scalar_feature("unknown_feature")][0][0]
        == 0
    )
    assert (
        model.weights()[model.get_index_for_scalar_feature("unknown_feature")][1][0]
        == 0
    )
    assert (
        model.weights()[model.get_index_for_scalar_feature("unknown_feature")][2][0]
        == 0
    )
    assert (
        model.weights()[model.get_index_for_scalar_feature("unknown_feature")][3][0]
        == 0
    )


# TODO: uncomment when implemented
# @pytest.mark.skip("Not yet implemented")
# def test_get_interacted_index() -> None:
#     model = vw.Workspace()

#     assert 2692 == model.get_index_for_interacted_feature(
#         [("a", None, " "), ("a", None, " ")]
#     )
#     assert 210827 == model.get_index_for_interacted_feature(
#         [("thing", None, "test"), ("val", "test", "another")]
#     )



import typing

from vowpal_wabbit_next import _core, Workspace, Example
from types import TracebackType

T = typing.TypeVar("T")


class DSJsonFormatParser:
    def __init__(self, workspace: Workspace[T]):
        """Parse VW DSJson format examples.

        Args:
            workspace (Workspace): Workspace object used to configure this parser
        """
        self._workspace = workspace

    def parse_json(self, text: str) -> typing.List[Example]:
        '''Parse a single json object in dsjson format.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, DSJsonFormatParser
            >>> workspace = Workspace(["--cb_explore_adf"])
            >>> parser = DSJsonFormatParser(workspace)
            >>> json_str = """
            ... {
            ...     "_label_cost": -1.0,
            ...     "_label_probability": 0.5,
            ...     "_label_Action": 2,
            ...     "_labelIndex": 1,
            ...     "a": [2, 1],
            ...     "c": {
            ...         "shared": { "f": "1" },
            ...         "_multi": [{ "action": { "f": "1" } }, { "action": { "f": "2" } }]
            ...     },
            ...     "p": [0.5, 0.5]
            ... }
            ... """
            >>> example = parser.parse_json(json_str)

        Args:
            text (str): JSON string of input

        Returns:
            typing.List[Example]: List of parsed examples
        '''
        return [
            Example(_existing_example=ex, _label_type=self._workspace.label_type)
            for ex in _core._parse_line_dsjson(self._workspace._workspace, text)
        ]


DSJsonFormatReaderT = typing.TypeVar("DSJsonFormatReaderT", bound="DSJsonFormatReader")


# takes a file and uses a context manager to generate based on the contents of the file
class DSJsonFormatReader:
    def __init__(self, workspace: Workspace[T], file: typing.TextIO):
        """Read VW DSJson format examples from the given text file. This reader always produces lists of examples.

        Examples:
            >>> from vowpal_wabbit_next import Workspace, DSJsonFormatReader
            >>> workspace = Workspace()
            >>> with open("data.txt", "r") as f:
            ...     with DSJsonFormatReader(workspace, f) as reader:
            ...         for example in reader:
            ...               workspace.predict_one(example)


        Args:
            workspace (Workspace): Workspace object used to configure this reader
            file (typing.BinaryIO): File to read from
        """
        self._parser = DSJsonFormatParser(workspace)
        self._workspace = workspace
        self._file = file
        if not self._workspace.multiline:
            raise ValueError("Must use a multiline Workspace for dsjson format")

    def __enter__(self: DSJsonFormatReaderT) -> DSJsonFormatReaderT:
        return self

    def __exit__(
        self,
        exc_type: typing.Optional[typing.Type[BaseException]],
        exc_value: typing.Optional[BaseException],
        traceback: typing.Optional[TracebackType],
    ) -> None:
        self._file.close()

    def __iter__(self) -> typing.Iterator[typing.List[Example]]:
        if self._workspace.multiline:
            for line in self._file:
                yield self._parser.parse_json(line.rstrip())
